#!/usr/bin/perl
#
# A simple e-voting CGI script
#
# $Id: run.pl,v 1.11 2011/09/17 09:22:11 dds Exp $
#
# Copyright (c) 2008-2011, Diomidis Spinellis
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#

# Character set
$charset = 'UTF-8';
# RADIUS server
$radius_server = '195.251.255.157';
# LDAP server
$ldap_server = 'ldap.aueb.gr';
# Radius shared secret
$radius_secret = 'SECRET';			# FIXME
# Administrator allowed to enter new questions
$admin = 'dds';
# Database passwords
$passwd_admin = 'SECRET';			# FIXME
$passwd_view = 'SECRET';			# FIXME
$passwd_vote = 'SECRET';			# FIXME

use CGI qw/:all/;
use DBI;
use Authen::Radius;
use Net::LDAP;

# Select a poll and its active status
$selpoll = q{
SELECT *,
    now() >= tstart and now() <= tend as active,
    now() > tend as ended
    FROM Polls
};

process_request();

sub
process_request
{
	if ($id = url_param('results')) {
		display_results($id);
	} elsif ($id = url_param('vote')) {
		display_options($id);
	} elsif (param('question')) {
		add_entry();
	} elsif (param('voteentry')) {
		add_vote();
	} elsif (url_param('new')) {
		new_entry();
	} else {
		display_entries();
	}
}

sub
display_results
{
	my($id) = @_;
	return unless ($dbh = database_connect("poll_view", $passwd_view));

	my $sth = $dbh->prepare($selpoll . q{WHERE id = ?;});
	if (!$sth->execute($id)) {
		user_error("Πρόβλημα κατά την ανάκτηση της ερώτησης.");
		return;
	}
	my $vote = $sth->fetchrow_hashref;
	if (!$vote->{'ended'}) {
		user_error("Η συγκεκριμένη ψηφοφορία δεν έχει ακόμα ολοκληρωθεί.");
		return;
	}
	my $ansh = $dbh->prepare(q{SELECT ordinal, answer, count FROM Options INNER JOIN Tallies ON Options.Id = Tallies.OptionId WHERE pollId = ? ORDER BY ordinal;});
	if (!$ansh->execute($id)) {
		user_error("Πρόβλημα κατά την ανάκτηση των απαντήσεων.");
		return;
	}
	start_page("Αποτελέσματα");
	print "<p>$vote->{'comment'}</p>\n";
	print "<p>$vote->{'question'}</p>\n";
	print "<table border='1'>\n";
	print "<tr><th>Ψήφοι</th><th>Απάντηση</th></tr>\n";
	while (my $opt = $ansh->fetchrow_hashref) {
		print "<tr><td align='right'>$opt->{'count'}</td><td>$opt->{'answer'}</td></tr>\n";
	}
	print "</table><p />\n";
	my ($sum) = $dbh->selectrow_array(qq{SELECT SUM(count) FROM Tallies INNER JOIN Options ON Tallies.OptionId = Options.Id WHERE pollId = ?;}, undef, $id);
	print "<p>Σύνολο ψήφων: $sum</p>\n";
	print "<p>Η ψηφοφορία δημιουργήθηκε στις $vote->{'posted'}.</p>\n";
	print "<p>Η ψηφοφορία άρχισε στις $vote->{'tstart'}.</p>\n";
	print "<p>Η ψηφοφορία έληξε στις $vote->{'tend'}.</p>\n";
	return_link();
	end_page();
}

sub
display_entries
{
	return unless ($dbh = database_connect("poll_view", $passwd_view));
	my $sth = $dbh->prepare($selpoll . q{ORDER BY posted DESC;});
	if (!$sth->execute()) {
		user_error("Πρόβλημα κατά την ανάκτηση των ψηφοφοριών.");
		return;
	}
	start_page("Ψηφοφορίες");
	print '<ul>';
	while (my $vote = $sth->fetchrow_hashref) {
		print "<li>$vote->{question}<br>\n";
		print qq(<a href="https://e-democracy.aueb.gr/vote?vote=$vote->{id}">Ψηφίστε</a> &mdash; \n) if ($vote->{active});
		print qq(<a href="http://e-democracy.aueb.gr/vote?results=$vote->{id}">Δείτε τα αποτελέσματα</a>\n) if($vote->{ended});
		print "($vote->{tstart}";
		print " έως $vote->{tend}";
		print ")\n";
		print "</li>\n";
	}
	print '</ul>';
	end_page();
}

sub
new_entry
{
	my($id) = @_;
	start_page("Νέα ψηφοφορία");
	print startform(-enctype=>'multipart/form-data');
	print
		'<p /><h2>Αντικείμενο της ψηφοφορίας</h2>',
		'<table>',
		'<tr><td>', 'Περιγραφή', '</td><td>',
		textarea(-name => 'comment',
			-default => '',
			-columns => 60,
			-rows => 10
			-maxlength => 1024,
			-override => 1),
		'</td></tr>',
		'<tr><td>', 'Ημερομηνία έναρξης', '</td><td>',
		textfield(-name => 'tstart',
			-default => '',
			-size => 20,
			-maxlength => 10,
			-override => 1),
		'</td></tr>',
		'<tr><td valign="top">', 'Ημερομηνία λήξης', '</td><td>',
		textfield(-name => 'tend',
			-default => '',
			-size => 20,
			-maxlength => 10,
			-override => 1),
		'<br />ΕΕΕΕ-ΜΜ-ΗΗ (π.χ. 2008-05-13)<br />',
		'</td></tr>',
		'<tr><td>', 'Ερώτηση', '</td><td>',
		textfield(-name => 'question',
			-default => '',
			-size => 80,
			-maxlength => 256,
			-override => 1),
		'</td></tr>';
	for (my $i = 1; $i < 10; $i++) {
		print
			'<tr><td>', "Απάντηση $i", '</td><td>',
			textfield(-name => "answer_$i",
				-default => '',
				-size => 80,
				-maxlength => 256,
				-override => 1),
			"</td></tr>\n";
	}
	print
		'<tr><td valign="top">', 'Κοινό', '</td><td>',
		textfield(-name => 'users',
			-default => '',
			-size => 80,
			-maxlength => 256,
			-override => 1),
		'<br />Αφήστε το πεδίο κενό για όλα τα μέλη της πανεπιστημιακής <br />',
		'κοινότητας ή συμπληρώστε "Μέλος ΔΕΠ", για να είναι η συγκεκριμένη <br />',
		'ψηφοφορία ανοικτή μόνο σε μέλη ΔΕΠ.',
		'</td></tr>';
	print '</table>';
	authentication_fields();
	print endform;
	end_page();
}

# Return true if OK false on error
#
sub
authenticate
{
	my ($users) = @_;
	my $uname = param('uname');
	my $passwd = param('passwd');
	$uname =~ s/[^a-zA-Z0-9_-]//g;

	if (length($users) > 0) {
		# LDAP: supports groups, but doesn't include students
		my $ldap = Net::LDAP->new($ldap_server, scheme => 'ldaps');
		if (!$ldap) {
			user_error("Τεχνικό πρόβλημα στην αυθεντικοποίηση LDAP: $@");
			return 0;
		}
		my $mesg = $ldap->bind("auebUserName=$uname,ou=People,dc=aueb,dc=gr", password => $passwd);
		if ($mesg->code) {
			user_error('Η αυθεντικοποίηση του χρήστη απέτυχε: ' . $mesg->error_text);
			return 0;
		}
		my $result = $ldap->search (base   => "ou=People, dc=aueb,dc=gr",
		    filter => "auebUserName=$uname" );
		if ($result->code) {
			user_error('Η ανάκτηση δεδομένων του χρήστη απέτυχε: ' . $mesg->error_text);
			return 0;
		}
		my @entries = $result->entries;
		my $entry = $entries[0];
		my @titles= $entry->get_value('title');
		if ($titles[0] !~ m/^($users)$/) {
			user_error("Ο χρήστης $uname δεν είναι μέλος της ομάδας $users");
			return 0;
		}
		return 1;
	} else {
		# RADIUS: No support for groups, but includes students
		my $r = new Authen::Radius(Host => $radius_server, Secret => $radius_secret);
		if ($r->check_pwd(param('uname'), param('passwd'))) {
			return 1;
		} else {
			user_error('Η αυθεντικοποίηση του χρήστη απέτυχε.');
			return 0;
		}
	}
}

sub
authentication_fields
{
	print
		'<p /><h2>Αυθεντικοποίηση</h2>',
		'<table>',
		'<tr><td>', 'Κωδικός χρήστη', '</td><td>',
		textfield(-name => 'uname',
			-default => '',
			-size => 10,
			-maxlength => 30,
			-override => 1),
		'</td></tr>',
		'<tr><td>', 'Συνθηματικό', '</td><td>',
		password_field(-name => 'passwd',
			-default => '',
			-size => 10,
			-maxlength => 30,
			-override => 1),
		'</td></tr></table>',
		submit(-value=>'Εισαγωγή', -name=>"login");
}

# Connect to the database with the specified credentials
# Return handle if OK undef on error
sub
database_connect
{
	my($uname, $passwd) = @_;

	my $dbh = DBI->connect("dbi:mysql:dbname=poll",$uname, $passwd);
	if ($dbh) {
		return $dbh;
	} else {
		user_error("Database connection to failed");
		return undef;
	}
}

sub
add_vote
{
	my $user;
	return unless ($dbh = database_connect("poll_vote", $passwd_vote));
	# Ensure that id and answer specify a valid answer
	if (!defined(param('answer'))) {
		user_error("Δεν έχετε επιλέξει απάντηση.");
		return;
	}
	if (!defined(param('id'))) {
		user_error("Δεν έχει οριστεί το θέμα.");
		return;
	}
	my $id = param('id');
	my $sth = $dbh->prepare($selpoll . q{WHERE id = ?});
	if (!$sth->execute($id)) {
		user_error("Πρόβλημα κατά την ανάκτηση της ερώτησης.");
		return;
	}
	my $poll = $sth->fetchrow_hashref;
	return unless (authenticate($poll->{'users'}));

	# Verify that the poll is active
	if (!$poll->{'active'}) {
		user_error("Η συγκεκριμένη ψηφοφορία δεν είναι ενεργή.");
		return;
	}

	# Verify that the answer is allowed
	$sth = $dbh->prepare(qq{SELECT Polls.id as id, Options.id as optionId, ordinal FROM Options LEFT JOIN Polls on Options.pollId = Polls.id WHERE Polls.id = ? AND Options.ordinal = ?});
	if (!$sth->execute($id, param('answer'))) {
		user_error("Πρόβλημα κατά την ανάκτηση της απάντησης.");
		return;
	}
	my $opt = $sth->fetchrow_hashref;
	if ($opt->{'id'} != $id || $opt->{'ordinal'} != param('answer')) {
		user_error("Λανθασμένος κωδικός θέματος ή απάντησης.");
		return;
	}
	my $optionId = $opt->{'optionId'};

	# Verify that the user hasn't voted by adding an entry into votes
	my ($sec, $min, $hour, $mday, $mon, $year) = gmtime(time);
	my $timestamp = sprintf("%04d-%02d-%02d %02d:%02d:%02d", $year + 1900, $mon + 1, $mday, $hour, $min, $sec);
	if (!$dbh->do('INSERT INTO Votes(name, pollId, ip, timestamp) VALUES(?, ?, ?, ?)', undef, (param('uname'), $id, $ENV{'REMOTE_ADDR'}, $timestamp))) {
		my $sth2 = $dbh->prepare(q{SELECT ip, timestamp FROM Votes WHERE name = ? and pollId = ?;});
		if (!$sth2->execute(param('uname'), $id)) {
			user_error("Πρόβλημα κατά την ανάκτηση των δεδομένων της προηγούμενης ψήφου.");
			return;
		}
		my $last_vote = $sth2->fetchrow_hashref;
		user_error("Δεν επιτρέπεται δεύτερη ψήφος από το ίδιο άτομο.  " .
			"(Η προηγούμενη ψήφος ρίφθηκε από τη διεύθυνση IP $last_vote->{'ip'} στις $last_vote->{'timestamp'} UTC.)");
		return;
	}
	# Update the tally
	if (!$dbh->do('UPDATE Tallies SET count = count + 1 WHERE optionId = ?', undef, ($optionId))) {
		user_error("Δεν ήταν δυνατή η ενημέρωση του αποτελέσματος.");
		return;
	}
	start_page("Η ψήφος σας καταχωρήθηκε");
	end_page();
	return;
}

sub
add_entry
{
	my $user;
	if (($user = param('uname')) ne $admin) {
		user_error("Ο χρήστης $user δεν έχει δικαίωμα εισαγωγής ερωτήσεων.");
		return;
	}
	return unless (authenticate());
	return unless ($dbh = database_connect("poll_admin", $passwd_admin));
	if (!$dbh->do('INSERT INTO Polls(question, comment, users, posted, tstart, tend) VALUES(?, ?, ?, now(), ?, ?)', undef, (param('question'), param('comment'), param('users'), param('tstart') . ' 00:00:00', param('tend') . ' 23:59:59'))) {
		user_error("Πρόβλημα κατά την εισαγωγή της ερώτησης.");
		return;
	}
	my ($pollId) = $dbh->selectrow_array(qq{SELECT LAST_INSERT_ID();});

	for (my $i = 1; $i < 10; $i++) {
		if (my $answer = param("answer_$i")) {
			if (!$dbh->do('INSERT INTO Options(ordinal, pollId, answer) VALUES(?, ?, ?)', undef, ($i, $pollId, $answer))) {
				user_error("Πρόβλημα κατά την εισαγωγή της απάντησης $i.");
				return;
			}
			if (!$dbh->do('INSERT INTO Tallies(optionId, count) VALUES((SELECT LAST_INSERT_ID()), 0)', undef, undef)) {
				user_error("Πρόβλημα κατά το μηδενισμό του αποτελέσματος.");
				return;
			}
		}
	}
	start_page("Αποτελέσματα εισαγωγής");
	print "<p>Η εισαγωγή της ερώτησης $pollId έγινε κανονικά.</p>\n";
	return_link();
	end_page();
}

sub
user_error
{
	my($msg) = @_;
	start_page("Λάθος");
	print "<p>$msg</p>\n";
	end_page();
}


sub
display_options
{
	my($id) = @_;
	return unless ($dbh = database_connect("poll_view", $passwd_view));
	my $sth = $dbh->prepare(q{SELECT question, comment, posted FROM Polls WHERE id = ?;});
	if (!$sth->execute($id)) {
		user_error("Πρόβλημα κατά την ανάκτηση της ερώτησης.");
		return;
	}
	my $ansh = $dbh->prepare(q{SELECT id, ordinal, answer FROM Options WHERE pollId = ? ORDER BY ordinal;});
	if (!$ansh->execute($id)) {
		user_error("Πρόβλημα κατά την ανάκτηση των απαντήσεων.");
		return;
	}
	start_page("Ψηφίστε");
	my $vote = $sth->fetchrow_hashref;
	print "<p>$vote->{'comment'}</p>\n";
	print "<p>$vote->{question}</p>\n";
	my(@values, %labels);
	while (my $opt = $ansh->fetchrow_hashref) {
		push(@values, $opt->{'ordinal'});
		$labels{$opt->{'ordinal'}} = $opt->{'answer'};
	}
	print startform(-enctype=>'multipart/form-data', -action=>'/vote');
	print
		'<p>', radio_group(-name=>'answer',
		-values=>\@values,
		-default=>'-',
		-linebreak=>'true',
		-labels=>\%labels),
		hidden('id',"$id"),
		hidden('voteentry','1'),
		"</p>\n";
	authentication_fields();
	print endform;
	end_page();
}

sub
start_page
{
	my($title) = @_;

	print header(-type => 'text/html', -charset => $charset);

	print q{<?xml version="1.0"?>
<html xmlns="http://www.w3.org/1999/xhtml">
<head>
    <title>e-democracy - Οικονομικό Πανεπιστήμιο Αθηνών</title>
    <meta name="generator" content="$Id: run.pl,v 1.11 2011/09/17 09:22:11 dds Exp $" />
    <meta http-equiv="Content-Type" content="text/html; charset=} . $charset . q{" />
    <meta name="Description" content="Υπηρεσία e-democracy του Οικονομικού Πανεπιστημίου Αθηνών" />
    <link href="css/default.css" rel="stylesheet" type="text/css" media="all" />
</head>

<body class="main">
<table id="tableWhiteBorder" align="center" border="0" cellpadding="0" cellspacing="0" width="800">

<!-- BEGIN TOPBAR inclusion -->
	<tr><td colspan="2" class="pagerDarkTop" height="10">
		<a href="http://e-democracy.aueb.gr/faq.html">Πληροφορίες &mdash; FAQ</a>
	</td></tr>
        <tr><td>
            <table border="0" cellpadding="0" cellspacing="0" width="100%">
                <tr><td id="siteName">
                   <div class="gb">
ΟΙΚΟΝΟΜΙΚΟ ΠΑΝΕΠΙΣΤΗΜΙΟ ΑΘΗΝΩΝ <br>
ATHENS UNIVERSITY OF ECONOMICS AND BUSINESS
<br clear=left>
e-democracy
		   </div>
		</td></tr>

	    </table>
	</td></tr>
<!-- END TOPBAR -->

	<tr><td>
};
	print qq{
  		<h1>$title</h1>
   		<div id="content_text">
};
}

sub
end_page
{
	print q{</div>

	</td></tr>

<!-- BEGIN BOTTOM BAR inclusion-->
        <tr><td colspan="2">
          <table id="pager" valign="top" border="0" cellpadding="0" cellspacing="0" width="100%">
            <tr><td class="pager">
                Οικονομικό Πανεπιστήμιο Αθηνών, Πατησίων 76, ΤΚ 104 34 Αθήνα
<br /> Tηλεφωνικό Κέντρο: +302108203 911, E-mail: <a href="mailto:webmaster@aueb.gr">webmaster@aueb.gr</a>
            </td></tr>
            <tr><td class="pagerDark" colspan="2" height="10">© Copyright 1996 - 2008 | Athens University of Economics and Business
	    </td></tr>
          </table>
        </td></tr>
<!-- END BOTTOM BAR inclusion -->

</table>
</body>
</html>
};
}

sub
return_link
{
	print qq{<p><a href="http://e-democracy.aueb.gr">Επιστροφή στην κεντρική σελίδα</a></p>\n};
}

# vim: encoding=utf-8
